import pytest
from basis_set_exchange import misc

@pytest.mark.parametrize("elements, expected", 
                         [ ([1], "H"),
                           ([1,2], "H,He"),
                           ([1,10], "H,Ne"),
                           ([1,2,3,11,23,24], "H-Li,Na,V,Cr")])
def test_compact_string(elements, expected):
    compacted = misc.compact_elements(elements)
    assert compacted == expected

    expanded = misc.expand_elements(compacted)
    assert expanded == elements



'''
Conversion of basis sets to Gaussian format
'''

from .g94 import write_g94


def write_psi4(basis):
    '''Converts a basis set to Psi4 format

    Psi4 uses the same output as gaussian94, except
    that the first line must be cartesian/spherical,
    and it prefers to have a starting asterisks

    The cartesian/spherical line is added later, since it must
    be the first non-blank line.
    '''

    s = '****\n'
    s += write_g94(basis)
    return s



'''
Functions for helping curate BSE basis set data
'''

from .readers import read_formatted_basis
from .metadata import create_metadata_file
from .compare_ref import compare_basis_against_ref, replace_basis_data
from .printing import (print_electron_shell, print_ecp_pot, print_element, print_component_basis, print_element_basis,
                       print_table_basis)
from .compare import (compare_electron_shells, electron_shells_are_subset, electron_shells_are_equal, compare_ecp_pots,
                      ecp_pots_are_subset, ecp_pots_are_equal, compare_elements, compare_basis, shells_difference,
                      potentials_difference, subtract_electron_shells)



"""
Tests for the BSE main API
"""

import random

import basis_set_exchange as bse
import pytest

from .common_testvars import *

# Use random for getting sets of elements
random.seed(rand_seed, version=2)

@pytest.mark.slow
@pytest.mark.parametrize('basis_name', bs_names)
@pytest.mark.parametrize('fmt', bs_formats)
@pytest.mark.parametrize('unc_gen', true_false)
@pytest.mark.parametrize('unc_seg', true_false)
@pytest.mark.parametrize('unc_spdf', true_false)
@pytest.mark.parametrize('make_gen', true_false)
@pytest.mark.parametrize('opt_gen', true_false)
def test_slow_get_basis_1(basis_name, fmt, unc_gen, unc_seg, unc_spdf, make_gen, opt_gen):
    """Tests getting all basis sets in all formats
       and with every combination of option

       Also tests memoization
    """

    this_metadata = bs_metadata[basis_name]
    for ver in this_metadata['versions'].keys():
        bs1 = bse.get_basis(basis_name, fmt=fmt, version=ver,
                            uncontract_general=unc_gen,
                            uncontract_segmented=unc_seg,
                            uncontract_spdf=unc_spdf,
                            make_general=make_gen,
                            optimize_general=opt_gen,
                            header=False)

        bs2 = bse.get_basis(basis_name, fmt=fmt, version=ver,
                            uncontract_general=unc_gen,
                            uncontract_segmented=unc_seg,
                            uncontract_spdf=unc_spdf,
                            make_general=make_gen,
                            optimize_general=opt_gen,
                            header=False)

        assert bs1 == bs2



'''
Testing of a string containing bibtex entries

This tests by running latex/bibtex on a file containing the
string. Warnings are interpreted as failures (which catches
missing fields)
'''

import os
import shutil
import subprocess
import tempfile

# Other modules should query this variable
available = bool(shutil.which('bibtex')) and bool(shutil.which('latex'))

# Test latex file
test_txt = r"""
\documentclass{article}
\begin{document}
\nocite{*}
\bibliographystyle{plain}
\bibliography{bse_test}
\end{document}
"""


def validate_bibtex(bib_str):
    with tempfile.TemporaryDirectory(prefix='bsebibtextest') as d:
        texfile = os.path.join(d, 'bse.tex')
        bibfile = os.path.join(d, 'bse_test.bib')

        with open(texfile, 'w') as tf:
            tf.write(test_txt)

        with open(bibfile, 'w') as bf:
            bf.write(bib_str)

        res = subprocess.check_output(['latex', 'bse'], cwd=d)

        # Run bibtex. Then check return code and scan the output for "Warning"
        res = subprocess.run(
            ['bibtex', 'bse'], universal_newlines=True, cwd=d, stderr=subprocess.STDOUT, stdout=subprocess.PIPE)

        if "Warning" in res.stdout:
            raise RuntimeError("Warning found in output. Output is:\n" + res.stdout)



"""
Functions related to validating JSON files (including against schema)
"""

import jsonschema

from . import api
from . import fileio


def validate(file_type, file_path):
    """
    Validates a file against a schema

    Parameters
    ----------
    file_type : str
        Type of file to read. May be 'component', 'element', 'table', or 'references'
    file_path:
        Full path to the file to be validated

    Raises
    ------
    RuntimeError
        If the file_type is not valid (and/or a schema doesn't exist)
    ValidationError
        If the given file does not pass validation
    FileNotFoundError
        If the file given by file_path doesn't exist
    """

    _valid_schema = ['component', 'element', 'table', 'references']
    if file_type not in _valid_schema:
        raise RuntimeError("{} is not a valid file_type".format(file_type))

    to_validate = fileio._read_plain_json(file_path, False)
    schema = api.get_schema(file_type)
    jsonschema.validate(to_validate, schema)



"""
Compares basis sets with authoritative versions
"""

import os

import pytest
from basis_set_exchange import misc, curate
from .common_testvars import *

# Find the dir with all the bse files
_my_dir = os.path.dirname(os.path.abspath(__file__))
_auth_data_dir = os.path.join(_my_dir, 'sources')

# create a map of the sources dir
_basis_src_map = {}
for x in os.listdir(_auth_data_dir):
    # remove .fmt.bz2
    base,_ = os.path.splitext(x)
    base,_ = os.path.splitext(base)

    if base in _basis_src_map:
        raise RuntimeError("Duplicate basis set in authoritative sources: {}".format(base))

    _basis_src_map[base] = os.path.join(_auth_data_dir, x)

@pytest.mark.parametrize('basis_name_ver', list(_basis_src_map.keys()))
def test_authoritative(basis_name_ver):
    basis_name,ver = os.path.splitext(basis_name_ver)
    ver = ver[1:] # remove starting '.'
    basis_meta = bs_metadata[basis_name]

    ref_filename = _basis_src_map[basis_name_ver]

    if not basis_name in bs_names:
        raise RuntimeError("Source basis {} doesn't have a BSE basis".format(basis_name))

    assert curate.compare_basis_against_ref(basis_name, ref_filename, version=ver, uncontract_general=True) 
            



'''
Class/decorator for memoizing BSE functionality
'''

import functools
import pickle

# If set to True, memoization of some internal functions
# will be used. Generally safe to leave enabled - it
# won't use that much memory
memoize_enabled = True


class BSEMemoize:
    def __init__(self, f):
        self.__f = f
        self.__memo = {}
        functools.update_wrapper(self, f)

    def __call__(self, *args):
        if memoize_enabled is not True:
            return self.__f(*args)

        if args in self.__memo:
            return pickle.loads(self.__memo[args])

        ret = self.__f(*args)
        self.__memo[args] = pickle.dumps(ret)
        return ret



'''
Main interface to Basis Set Exchange functionality
'''

import datetime
import os
import textwrap

from collections import OrderedDict

from . import compose
from . import converters
from . import fileio
from . import lut
from . import manip
from . import memo
from . import notes
from . import refconverters
from . import references
from . import misc

# Determine the path to the data directory that is part
# of this installation
_my_dir = os.path.dirname(os.path.abspath(__file__))
_default_data_dir = os.path.join(_my_dir, 'data')
_default_schema_dir = os.path.join(_my_dir, 'schema')

# Main URL of the project
_main_url = 'http://bse.pnl.gov'


def _convert_element_list(elements):
    '''Convert a list of elements to an internal list

    A list of elements can contain both integers and strings.
    This converts the list entirely to strings containing integer
    Z-numbers
    '''

    ret = []
    for el in elements:
        if isinstance(el, int):
            ret.append(str(el))
        elif isinstance(el, str) and not el.isdecimal():
            ret.append(str(lut.element_Z_from_sym(el)))
        else:
            ret.append(el)

    return ret


def _convert_version(ver):
    '''Convert a version to a string

    Versions can be either integers or strings.
    '''

    if isinstance(ver, int):
        return str(ver)
    else:
        return ver


def _get_basis_metadata(name, data_dir):
    '''Get metadata for a single basis set

    If the basis doesn't exist, an exception is raised
    '''

    # Transform the name into an internal representation
    tr_name = misc.transform_basis_name(name)

    # Get the metadata for all basis sets
    metadata = get_metadata(data_dir)

    if not tr_name in metadata:
        raise KeyError("Basis set {} does not exist".format(name))

    return metadata[tr_name]


def _header_string(basis_dict):
    '''Creates a header with information about a basis set

    Information includes description, revision, etc, but not references
    '''

    dt = datetime.datetime.utcnow()
    timestamp = dt.strftime('%Y-%m-%d %H:%M:%S UTC')

    tw = textwrap.TextWrapper(initial_indent='', subsequent_indent=' ' * 20)

    header = '-' * 70 + '\n'
    header += ' Basis Set Exchange\n'
    header += ' ' + _main_url + '\n'
    header += ' Accessed ' + timestamp + '\n'
    header += '-' * 70 + '\n'
    header += '   Basis set: ' + basis_dict['basis_set_name'] + '\n'
    header += tw.fill(' Description: ' + basis_dict['basis_set_description']) + '\n'
    header += '        Role: ' + basis_dict['basis_set_role'] + '\n'
    header += tw.fill('     Version: {}  ({})'.format(basis_dict['basis_set_version'],
                                                      basis_dict['basis_set_revision_description'])) + '\n'
    header += '-' * 70 + '\n'

    return header


def get_basis(name,
              elements=None,
              version=None,
              fmt=None,
              uncontract_general=False,
              uncontract_spdf=False,
              uncontract_segmented=False,
              make_general=False,
              optimize_general=False,
              data_dir=None,
              header=True):
    '''Obtain a basis set

    This is the main function for getting basis set information.
    This function reads in all the basis data and returns it either
    as a string or as a python dictionary.

    Parameters
    ----------
    name : str
        Name of the basis set. This is not case sensitive.
    elements : list
        List of elements that you want the basis set for. By default,
        all elements for which the basis set is defined are included.
        Elements can be specified by Z-number (int or str) or by symbol (str)
    version : int or str
        Obtain a specific version of this basis set. By default,
        the latest version is returned.
    fmt: str
        What format to return the basis set as. By defaut,
        basis set information is returned as a python dictionary. Otherwise,
        if a format is specified, a string is returned.
        Use :func:`bse.api.get_formats` to obtain the available formats.
    uncontract_general : bool
        If True, remove general contractions by duplicating the set
        of primitive exponents with each vector of coefficients.
        Primitives with zero coefficient are removed, as are duplicate shells.
    uncontract_spdf : bool
        If True, remove general contractions with combined angular momentum (sp, spd, etc)
        by duplicating the set of primitive exponents with each vector of coefficients.
        Primitives with zero coefficient are removed, as are duplicate shells.
    uncontract_segmented : bool
        If True, remove segmented contractions by duplicating each primitive into new shells.
        Each coefficient is set to 1.0
    make_general : bool
        If True, make the basis set as generally-contracted as possible. There will be one
        shell per angular momentum (for each element)
    optimize_general : bool
        Optimize by removing general contractions that contain uncontracted
        functions (see :func:`bse.manip.optimize_general`)
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    bs_data = _get_basis_metadata(name, data_dir)

    # If version is not specified, use the latest
    if version is None:
        version = bs_data['latest_version']
    else:
        version = _convert_version(version)

    # Compose the entire basis set (all elements)
    file_relpath = bs_data['versions'][version]['file_relpath']
    basis_dict = compose.compose_table_basis(file_relpath, data_dir)

    # Handle optional arguments
    if elements is not None:
        # Convert to purely a list of strings that represent integers
        elements = _convert_element_list(elements)

        bs_elements = basis_dict['basis_set_elements']

        # Are elements part of this basis set?
        for el in elements:
            if not el in bs_elements:
                raise KeyError("Element {} not found in basis {}".format(el, name))

            # Set to only the elements we want
            basis_dict['basis_set_elements'] = {k: v for k, v in bs_elements.items() if k in elements}

    if optimize_general:
        basis_dict = manip.optimize_general(basis_dict)
    if uncontract_general:
        basis_dict = manip.uncontract_general(basis_dict)
    if uncontract_spdf:
        basis_dict = manip.uncontract_spdf(basis_dict)
    if uncontract_segmented:
        basis_dict = manip.uncontract_segmented(basis_dict)
    if make_general:
        basis_dict = manip.make_general(basis_dict)

    # If fmt is not specified, return as a python dict
    if fmt is None:
        return basis_dict

    # make converters case insensitive
    if header:
        header_str = _header_string(basis_dict)
    else:
        header_str = None

    return converters.convert_basis(basis_dict, fmt, header_str)


def lookup_basis_by_role(primary_basis, role, data_dir=None):
    '''Lookup the name of a basis set given a primary basis set and role
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    bs_data = _get_basis_metadata(primary_basis, data_dir)
    auxdata = bs_data['auxiliaries']

    if not role in auxdata:
        raise RuntimeError("Role {} doesn't exist for {}".format(role, primary_basis))

    return auxdata[role]


@memo.BSEMemoize
def get_metadata(data_dir=None):
    '''Obtain the metadata for all basis sets

    The metadata includes information such as the display name of the basis set,
    its versions, and what elements are included in the basis set

    The data is read from the METADATA.json file in the `data_dir` directory.

    Parameters
    ----------
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    metadata_file = os.path.join(data_dir, "METADATA.json")
    return fileio.read_metadata(metadata_file)


@memo.BSEMemoize
def get_reference_data(data_dir=None):
    '''Obtain information for all stored references

    This is a nested dictionary with all the data for all the references

    The reference data is read from the REFERENCES.json file in the given
    `data_dir` directory.
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    reffile_path = os.path.join(data_dir, 'REFERENCES.json')

    return fileio.read_references(reffile_path)


def get_all_basis_names(data_dir=None):
    '''Obtain a list of all basis set names

    The returned list is the internal representation of the basis set name.

    The data is read from the METADATA.json file in the data directory.

    Parameters
    ----------
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    return sorted(list(get_metadata(data_dir).keys()))


def get_references(name, elements=None, version=None, fmt=None, data_dir=None):
    '''Get the references/citations for a basis set

    Parameters
    ----------
    name : str
        Name of the basis set. This is not case sensitive.
    elements : list
        List of element numbers that you want the basis set for. By default,
        all elements for which the basis set is defined are included.
    version : int
        Obtain a specific version of this basis set. By default,
        the latest version is returned.
    fmt: str
        What format to return the basis set as. By defaut,
        basis set information is returned as a list of dictionaries. Use
        get_reference_formats() to obtain the available formats.
    data_dir : str
        Data directory with all the basis set information. By default,
        it is in the 'data' subdirectory of this project.
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    basis_dict = get_basis(name, elements=elements, version=version, data_dir=data_dir)

    all_ref_data = get_reference_data(data_dir)
    ref_data = references.compact_references(basis_dict, all_ref_data)

    if fmt is None:
        return ref_data

    return refconverters.convert_references(ref_data, fmt)


def get_basis_family(name, data_dir=None):
    '''Lookup a family by a basis set name
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    bs_data = _get_basis_metadata(name, data_dir)
    return bs_data['family']


@memo.BSEMemoize
def get_family_notes(family, data_dir=None):
    '''Return a string representing the notes about a basis set family
    
    If the notes are not found, a string saying so is returned
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    file_name = 'NOTES.' + family.lower()
    file_path = os.path.join(data_dir, file_name)

    notes_str = fileio.read_notes_file(file_path)
    if notes_str is None:
        notes_str = "Notes are not available for the {} family".format(family)

    ref_data = get_reference_data(data_dir)
    return notes.process_notes(notes_str, ref_data)


@memo.BSEMemoize
def get_basis_notes(name, data_dir=None):
    '''Return a string representing the notes about a specific basis set
    
    If the notes are not found, a string saying so is returned
    '''

    data_dir = _default_data_dir if data_dir is None else data_dir
    bs_data = _get_basis_metadata(name, data_dir)

    # the notes file is the same as the base file name, with a .notes extension
    latest_ver = bs_data['latest_version']
    bs_ver_data = bs_data['versions'][latest_ver]
    rel_path = bs_ver_data['file_relpath']
    filebase = os.path.splitext(rel_path)[0]  # remove .json
    filebase = os.path.splitext(filebase)[0]  # remove .tablejson
    filebase = os.path.splitext(filebase)[0]  # remove .[version]
    file_path = os.path.join(data_dir, filebase + '.notes')

    notes_str = fileio.read_notes_file(file_path)
    if notes_str is None:
        notes_str = "Notes are not available for the {} basis".format(bs_data['display_name'])

    ref_data = get_reference_data(data_dir)
    return notes.process_notes(notes_str, ref_data)


def get_schema(schema_type):
    '''Get a schema that can validate BSE JSON files

       The schema_type represents the type of BSE JSON file to be validated,
       and can be 'component', 'element', 'table', or 'references'.
    '''

    schema_file = "{}-schema.json".format(schema_type)
    file_path = os.path.join(_default_schema_dir, schema_file)

    if not os.path.isfile(file_path):
        raise RuntimeError('Schema file \'{}\' does not exist, is not readable, or is not a file'.format(file_path))

    return fileio.read_schema(file_path)


def get_formats():
    '''Return information about the basis set formats available

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`get_basis()`
    '''

    return converters.get_formats()


def get_reference_formats():
    '''Return information about the reference/citation formats available

    The returned data is a map of format to display name. The format
    can be passed as the fmt argument to :func:`get_references`
    '''
    return refconverters.get_formats()


def get_roles():
    '''Return information about the available basis set roles available

    The returned data is a map of role to display name. The format
    can be passed as the role argument to fmt argument to :func:`lookup_basis_by_role`
    '''

    return OrderedDict((('orbital', 'Orbital basis'),
                        ('jfit', 'J-fitting'),
                        ('jkfit', 'JK-fitting'),
                        ('rifit', 'RI-fitting')))



"""
Tests BSE curation functions
"""

import os
import tempfile

import pytest
from basis_set_exchange import api, curate, manip
from .common_testvars import bs_names_sample

roundtrip_formats = ['turbomole', 'gaussian94', 'nwchem']


@pytest.mark.parametrize('basis', bs_names_sample)
@pytest.mark.parametrize('fmt', roundtrip_formats)
def test_curate_roundtrip(basis, fmt):

    # Many formats have limitations on general contractions
    if fmt == 'gaussian94':
        uncontract_general = True
        uncontract_spdf = 1
    if fmt == 'turbomole':
        uncontract_general = True
        uncontract_spdf = 0
    if fmt == 'nwchem':
        uncontract_general = False
        uncontract_spdf = 1

    bse_formatted = api.get_basis(basis, fmt=fmt)
    bse_dict = api.get_basis(basis, uncontract_general=uncontract_general)
    bse_dict = manip.uncontract_spdf(bse_dict, uncontract_spdf)

    outfile = tempfile.NamedTemporaryFile(mode='w', delete=False)
    outfile_path = outfile.name
    outfile.write(bse_formatted)
    outfile.close()

    test_dict = curate.read_formatted_basis(outfile_path, fmt)
    os.remove(outfile_path)

    test_dict = manip.sort_basis(test_dict)
    bse_dict = manip.sort_basis(bse_dict)

    # Compare, ignoring metadata (not stored in most formats)
    assert curate.compare_basis(bse_dict, test_dict, rel_tol=0.0)



"""
Tests for lookups of elemental data
"""

import pytest
from basis_set_exchange import lut

_all_elements = list(range(1,119))
_all_am = list(range(1, 13))

@pytest.mark.parametrize('Z', _all_elements)
def test_element_data(Z):
    # Cycle through the elements and check that 
    # the info returned from the functions is consistent
    data = lut.element_data_from_Z(Z)
    assert data[1] == Z

    assert data == lut.element_data_from_sym(data[0])
    assert data == lut.element_data_from_name(data[2])

    assert data[0] == lut.element_sym_from_Z(Z)
    assert data[2] == lut.element_name_from_Z(Z)

    nsym = lut.element_sym_from_Z(Z, True)
    nname = lut.element_name_from_Z(Z, True)

    assert nsym[0] == data[0][0].upper()
    assert nname[0] == data[2][0].upper()


@pytest.mark.parametrize('am', _all_am)
def test_amchar(am):
    # Check that converting am characters, etc, is consistent
    s = lut.amint_to_char([am])
    assert am == lut.amchar_to_int(s)[0]

    combined = list(range(am+1))
    s = lut.amint_to_char(combined)
    assert combined == lut.amchar_to_int(s)



"""
Tests for sanity of auxiliary basis sets
"""

import basis_set_exchange as bse

from .common_testvars import *

def test_roles():
    """Make sure all basis sets have a role that is in api.get_roles()
    """

    roles = bse.get_roles()

    # All the roles in the metadata
    for k,v in bs_metadata.items():
        r = v['role']
        assert(r in roles)



'''
Read a basis set file in a given format
'''

import os
import bz2
from .turbomole import read_turbomole
from .g94 import read_g94
from .nwchem import read_nwchem
from .gbasis import read_gbasis

_type_readers = {
    'turbomole': {
        'reader': read_turbomole,
        'extension': '.tm',
    },
    'gaussian94': {
        'reader': read_g94,
        'extension': '.gbs'
    },
    'nwchem': {
        'reader': read_nwchem,
        'extension': '.nw'
    },
    'gbasis': {
        'reader': read_gbasis,
        'extension': '.gbasis'
    }
}


def _fix_uncontracted(basis):
    '''
    Forces the contraction coefficient of uncontracted shells to 1.0
    '''

    for el in basis['basis_set_elements'].values():
        if 'element_electron_shells' not in el:
            continue

        for sh in el['element_electron_shells']:
            if len(sh['shell_coefficients']) == 1 and len(sh['shell_coefficients'][0]) == 1:
                sh['shell_coefficients'][0][0] = '1.0000000'

            # Some uncontracted shells don't have a coefficient
            if len(sh['shell_coefficients']) == 0:
                sh['shell_coefficients'].append(['1.0000000'])

    return basis


def read_formatted_basis(file_path, file_type=None):

    if not os.path.isfile(file_path):
        raise RuntimeError('Basis file path \'{}\' does not exist'.format(file_path))

    if file_type is None:
        for k, v in _type_readers.items():
            ext = v['extension']
            ext_bz2 = ext + '.bz2'
            if file_path.endswith(ext) or file_path.endswith(ext_bz2):
                file_type = k
                break

        else:
            raise RuntimeError("Unable to determine basis set format of '{}'".format(file_path))
    else:
        file_type = file_type.lower()

    if file_type not in _type_readers:
        raise RuntimeError("Unknown file type to read '{}'".format(file_type))

    fname = os.path.basename(file_path)

    # Handle compressed files
    if file_path.endswith('.bz2'):
        with bz2.open(file_path, 'rt') as f:
            basis_lines = [l.strip() for l in f.readlines()]
    else:
        with open(file_path, 'r') as f:
            basis_lines = [l.strip() for l in f.readlines()]

    data = _type_readers[file_type]['reader'](basis_lines, fname)
    return _fix_uncontracted(data)



"""
Tests for the BSE IO functions
"""

# Most functionality is covered under other tests.
# This tests the remainder

import os

import pytest
from basis_set_exchange import fileio, api

_data_dir = api._default_data_dir

@pytest.mark.parametrize('file_path', [
                              'cc-pVDZ.0.table.json',
                              'CRENBL.0.table.json',
                              'dunning/cc-pVDZ.1.element.json',
                              'crenb/CRENBL.0.element.json',
                              'dunning/cc-pVDZ_dunning1989a.1.json',
                              'crenb/CRENBL_ross1994a.0.json',
                              'crenb/CRENBL-ECP_ross1994a.0.json'
                         ])
def test_read_write_basis(file_path):
    # needed to be tested to make sure something isn't left
    # out of the sort lists, etc
    full_path = os.path.join(_data_dir, file_path)
    full_path_new = full_path + '.new'
    data = fileio.read_json_basis(full_path)
    fileio.write_json_basis(full_path_new, data)
    os.remove(full_path_new)


@pytest.mark.parametrize('file_path', [
                              'REFERENCES.json'
                         ])
def test_read_write_references(file_path):
    # needed to be tested to make sure something isn't left
    # out of the sort lists, etc
    full_path = os.path.join(_data_dir, file_path)
    full_path_new = full_path + '.new'
    data = fileio.read_references(full_path)
    fileio.write_references(full_path_new, data)
    os.remove(full_path_new)



from ... import lut


def read_g94(basis_lines, fname):
    '''Reads G94-formatted file data and converts it to a dictionary with the
       usual BSE fields

       Note that the gaussian format does not store all the fields we
       have, so some fields are left blank
    '''

    skipchars = '!'
    basis_lines = [l for l in basis_lines if l and not l[0] in skipchars]

    bs_data = {
        'molssi_bse_schema': {
            'schema_type': 'component',
            'schema_version': '0.1'
        },
        'basis_set_description': fname,
        'basis_set_references': [],
        'basis_set_elements': {}
    }

    i = 0
    if basis_lines[0] == '****':
        i += 1  # skip initial ****

    while i < len(basis_lines):
        line = basis_lines[i]
        elementsym = line.split()[0]

        # Some gaussian files have a dash before the element
        if elementsym[0] == '-':
            elementsym = elementsym[1:]

        element_Z = lut.element_Z_from_sym(elementsym)
        element_Z = str(element_Z)

        if not element_Z in bs_data['basis_set_elements']:
            bs_data['basis_set_elements'][element_Z] = {}

        element_data = bs_data['basis_set_elements'][element_Z]

        i += 1

        # Try to guess if this is an ecp
        # Electron basis almost always end in 1.0 (scale factor)
        # ECP lines would end in an integer, so isdecimal() = true for ecp
        if basis_lines[i].split()[-1].isdecimal():
            if not 'element_ecp' in element_data:
                element_data['element_ecp'] = []

            lsplt = basis_lines[i].split()
            maxam = int(lsplt[1])
            n_elec = int(lsplt[2])
            element_data['element_ecp_electrons'] = n_elec

            # Highest AM first, then the rest in order
            am_list = list(range(maxam + 1))
            am_list.insert(0, am_list.pop())

            i += 1

            for j in range(maxam + 1):
                i += 1  # Skip the 'title' block - unused according to gaussian docs
                n_entries = int(basis_lines[i])
                i += 1  # Skip title block

                shell_am = am_list[j]
                ecp_shell = {'potential_angular_momentum': [shell_am], 'potential_ecp_type': 'scalar'}
                rexponents = []
                gexponents = []
                coefficients = []

                for k in range(n_entries):
                    lsplt = basis_lines[i].split()
                    rexponents.append(int(lsplt[0]))
                    gexponents.append(lsplt[1])
                    coefficients.append(lsplt[2:])
                    i += 1

                ecp_shell['potential_r_exponents'] = rexponents
                ecp_shell['potential_gaussian_exponents'] = gexponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                ecp_shell['potential_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_ecp'].append(ecp_shell)
        else:
            if not 'element_electron_shells' in element_data:
                element_data['element_electron_shells'] = []

            while basis_lines[i] != '****':
                lsplt = basis_lines[i].split()
                shell_am = lut.amchar_to_int(lsplt[0])
                nprim = int(lsplt[1])

                shell = {
                    'shell_function_type': 'gto',
                    'shell_harmonic_type': 'spherical',
                    'shell_region': 'valence',
                    'shell_angular_momentum': shell_am
                }

                exponents = []
                coefficients = []

                i += 1
                for j in range(nprim):
                    line = basis_lines[i].replace('D', 'E')
                    line = line.replace('d', 'E')
                    lsplt = line.split()
                    exponents.append(lsplt[0])
                    coefficients.append(lsplt[1:])
                    i += 1

                shell['shell_exponents'] = exponents

                # We need to transpose the coefficient matrix
                # (we store a matrix with primitives being the column index and
                # general contraction being the row index)
                shell['shell_coefficients'] = list(map(list, zip(*coefficients)))

                element_data['element_electron_shells'].append(shell)

            i += 1

    return bs_data



'''
Miscellaneous helper functions
'''

from . import lut


def compact_elements(elements):
    """
    Create a string (with ranges) given a list of element numbers

    For example, [1, 2, 3, 6, 7, 8, 10] will return "H-Li,C-O,Ne"
   """

    if len(elements) == 0:
        return

    # We have to convert to integers for this function
    elements = [int(el) for el in elements]

    # Just to be safe, sort the list
    el = sorted(set(elements))

    ranges = []
    i = 0
    while i < len(el):
        start_el = el[i]
        end_el = start_el

        i += 1
        while i < len(el):
            if el[i] != end_el + 1:
                break

            end_el += 1
            i += 1

        if start_el == end_el:
            ranges.append([start_el])
        else:
            ranges.append([start_el, end_el])

    # Convert to elemental symbols
    range_strs = []
    for r in ranges:
        sym = lut.element_sym_from_Z(r[0], True)

        if len(r) == 1:
            range_strs.append(sym)
        elif len(r) == 2 and r[1] == r[0] + 1:
            sym2 = lut.element_sym_from_Z(r[1], True)
            range_strs.append(sym + "," + sym2)
        else:
            sym2 = lut.element_sym_from_Z(r[1], True)
            range_strs.append(sym + "-" + sym2)

    return ",".join(range_strs)


def expand_elements(compact_str):
    """
    Create a list of integers given a string of compacted elements

    This is the opposite of compact_elements

    For example, "H-Li,C-O,Ne" will return [1, 2, 3, 6, 7, 8, 10]
    """

    if len(compact_str) == 0:
        return

    # Split on commas
    tmp_list = compact_str.split(',')

    # Now go over each one and replace elements with ints
    el_list = []
    for el in tmp_list:
        if not '-' in el:
            el_list.append(lut.element_Z_from_sym(el))
        else:
            begin, end = el.split('-')
            begin = lut.element_Z_from_sym(begin)
            end = lut.element_Z_from_sym(end)
            el_list.extend(list(range(int(begin), int(end) + 1)))

    return el_list


def transform_basis_name(name):
    """
    Transforms the name of a basis set to an internal representation

    This makes comparison of basis set names easier by, for example,
    converting the name to all lower case.
    """

    return name.lower()



"""
Functions related to composing basis sets from individual components
"""

import os
from . import fileio, manip, memo

# If set to True, basis sets returned as python dictionaries
# will contain the path to a file where each shell/potential
# came from
debug_data_sources = False


def _whole_basis_harmonic(basis):
    '''
    Find whether an entire basis is cartesian, spherical, or if it is mixed

    May also return 'none' (as a string) if there are no electron shells (ie, ECP only)
    '''

    all_harm = set()

    for v in basis['basis_set_elements'].values():
        if not 'element_electron_shells' in v:
            return "none"
        for sh in v['element_electron_shells']:
            all_harm.add(sh['shell_harmonic_type'])

    if len(all_harm) > 1:
        return 'mixed'
    else:
        return all_harm.pop()


def compose_elemental_basis(file_relpath, data_dir):
    """
    Creates an 'elemental' basis from an elemental json file

    This function reads the info from the given file, and reads all the component
    basis set information from the files listed therein. It then composes all the
    information together into one 'elemental' basis dictionary
    """

    # Do a simple read of the json
    el_bs = fileio.read_json_basis(os.path.join(data_dir, file_relpath))

    # construct a list of all files to read
    component_files = set()
    for k, v in el_bs['basis_set_elements'].items():
        component_files.update(set(v['element_components']))

    # Read all the data from these files into a big dictionary
    component_map = {k: fileio.read_json_basis(os.path.join(data_dir, k)) for k in component_files}

    # If debugging, add file source info
    if debug_data_sources:
        for k, v in component_map.items():
            for el, el_data in v['basis_set_elements'].items():
                if 'element_electron_shells' in el_data:
                    for sh in el_data['element_electron_shells']:
                        sh['data_source'] = os.path.join(data_dir, k)
                if 'element_ecp' in el_data:
                    for sh in el_data['element_ecp']:
                        sh['data_source'] = os.path.join(data_dir, k)
                el_data['lkjasd'] = 'lakjsdlkajs;ldkjal;sdkjasd'

    # Broadcast the basis_set_references to each element
    # Use the basis_set_description for the reference description
    for k, v in component_map.items():
        for el, el_data in v['basis_set_elements'].items():
            el_data['element_references'] = [{
                'reference_description': v['basis_set_description'],
                'reference_keys': v['basis_set_references']
            }]

    # Compose on a per-element basis
    for k, v in el_bs['basis_set_elements'].items():

        components = v.pop('element_components')

        # all of the component data for this element
        el_comp_data = [component_map[c]['basis_set_elements'][k] for c in components]

        # merge all the data
        v = manip.merge_element_data(v, el_comp_data)
        el_bs['basis_set_elements'][k] = v

    return el_bs


@memo.BSEMemoize
def compose_table_basis(file_relpath, data_dir):
    """
    Creates a 'table' basis from an table json file

    This function reads the info from the given file, and reads all the elemental
    basis set information from the files listed therein. It then composes all the
    information together into one 'table' basis dictionary
    """

    # Do a simple read of the json
    file_path = os.path.join(data_dir, file_relpath)
    table_bs = fileio.read_json_basis(file_path)

    # construct a list of all elemental files to read
    component_files = set()
    for k, v in table_bs['basis_set_elements'].items():
        component_files.add(v['element_entry'])

    # Create a map of the elemental basis data
    element_map = {k: compose_elemental_basis(k, data_dir) for k in component_files}

    for k, v in table_bs['basis_set_elements'].items():
        entry = v['element_entry']
        data = element_map[entry]

        # Replace the basis set for this element with the one
        # from the elemental basis
        table_bs['basis_set_elements'][k] = data['basis_set_elements'][k]

    # Add the version to the dictionary
    file_base = os.path.basename(file_relpath)
    table_bs['basis_set_version'] = file_base.split('.')[-3]

    # Add whether the entire basis is spherical or cartesian
    table_bs['basis_set_harmonic_type'] = _whole_basis_harmonic(table_bs)

    return table_bs



'''
Conversion of references to bibtex format
'''

from ..misc import compact_elements


def _ref_bib(key, ref):
    '''Convert a single reference to bibtex format
    '''
    s = u''

    s += u'@{}{{{},\n'.format(ref['type'], key)

    entry_lines = []
    for k, v in ref.items():
        if k == 'type':
            continue

        # Handle authors/editors
        if k == 'authors':
            entry_lines.append(u'    author = {{{}}}'.format(' and '.join(v)))
        elif k == 'editors':
            entry_lines.append(u'    editor = {{{}}}'.format(' and '.join(v)))
        else:
            entry_lines.append(u'    {} = {{{}}}'.format(k, v))

    s += ',\n'.join(entry_lines)
    s += '\n}'

    return s


def write_bib(refs):
    '''Converts references to bibtex
    '''

    full_str = u''

    # First, write out the element, description -> key mapping
    # Also make a dict of unique reference to output
    unique_refs = {}

    for ref in refs:
        full_str += u'% {}\n'.format(compact_elements(ref['elements']))

        for ri in ref['reference_info']:
            full_str += u'%     {}\n'.format(ri['reference_description'])

            refdata = ri['reference_data']

            if len(refdata) == 0:
                full_str += u'%     (...no reference...)\n%\n'
            else:
                full_str += u'%         {}\n%\n'.format(' '.join(ri['reference_keys']))

            for k, r in refdata.items():
                unique_refs[k] = r

    full_str += u'\n\n'

    # Go through them sorted alphabetically by key
    for k, r in sorted(unique_refs.items(), key=lambda x: x[0]):
        full_str += u'{}\n\n'.format(_ref_bib(k, r))

    return full_str



import os
import setuptools
import versioneer

_my_dir = os.path.dirname(os.path.abspath(__file__))
_bse_dir = os.path.join(_my_dir, 'basis_set_exchange')
_readme_path = os.path.join(_my_dir, "README.md")

# Use the readme file for a description
with open(_readme_path, 'r') as readme_file:
    long_description = readme_file.read()

# Find the json files in the data dir and the schema dir
bse_package_data = []
for data_dir in 'data', 'schema', os.path.join('tests', 'sources'):
    for (path, _, filenames) in os.walk(os.path.join('basis_set_exchange', data_dir)):
        for filename in filenames:
            filepath = os.path.join(path, filename)
            bse_package_data.append(os.path.relpath(filepath, 'basis_set_exchange'))

if __name__ == "__main__":
    my_packages=setuptools.find_packages()

    setuptools.setup(
        name='basis_set_exchange',
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        description='The Quantum Chemistry Basis Set Exchange',
        long_description=long_description,
        long_description_content_type='text/markdown',
        author='The Molecular Sciences Software Institute',
        author_email='bpp4@vt.edu',
        url="https://github.com/MolSSI-BSE/basis_set_exchange",
        license='BSD-3C',
        packages=my_packages,
        install_requires=[
            'jsonschema',
        ],
        extras_require={
            'docs': [
                'sphinx',
                'sphinxcontrib-napoleon',
                'sphinx_rtd_theme',
                'numpydoc',
            ],
            'tests': [
                'pytest',
                'pytest-cov'
            ],
        },

        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3',
        ],

        package_data={'basis_set_exchange': bse_package_data},

        zip_safe=True,
    )



'''
Basis Set Exchange

Contains utilities for reading, writing, and converting
basis set information
'''

# Just import the basic user API
from .api import (get_basis, lookup_basis_by_role, get_metadata, get_reference_data, get_all_basis_names,
                  get_references, get_basis_family, get_family_notes, get_basis_notes, get_schema, get_formats,
                  get_reference_formats, get_roles)

# Handle versioneer
from ._version import get_versions
versions = get_versions()
__version__ = versions['version']
__git_revision__ = versions['full-revisionid']
del get_versions, versions



"""
Common basis set manipulations

This module contains functions for uncontracting and merging basis set
data, as well as some other small functions.
"""

import copy

from . import lut


def contraction_string(element):
    """
    Forms a string specifying the contractions for an element

    ie, (16s,10p) -> [4s,3p]
    """

    # Does not have electron shells (ECP only?)
    if 'element_electron_shells' not in element:
        return ""

    cont_map = dict()
    for sh in element['element_electron_shells']:
        nprim = len(sh['shell_exponents'])
        ngeneral = len(sh['shell_coefficients'])

        # is a combined general contraction (sp, spd, etc)
        is_spdf = len(sh['shell_angular_momentum']) > 1

        for am in sh['shell_angular_momentum']:
            # If this a general contraction (and not combined am), then use that
            ncont = ngeneral if not is_spdf else 1

            if am not in cont_map:
                cont_map[am] = (nprim, ncont)
            else:
                cont_map[am] = (cont_map[am][0] + nprim, cont_map[am][1] + ncont)

    primstr = ""
    contstr = ""
    for am in sorted(cont_map.keys()):
        nprim, ncont = cont_map[am]

        if am != 0:
            primstr += ','
            contstr += ','
        primstr += str(nprim) + lut.amint_to_char([am])
        contstr += str(ncont) + lut.amint_to_char([am])

    return "({}) -> [{}]".format(primstr, contstr)


def merge_element_data(dest, sources):
    """
    Merges the basis set data for an element from multiple sources
    into dest.

    The destination is not modified, and a (shallow) copy of dest is returned
    with the data from sources added.
    """

    ret = dest.copy()

    for s in sources:
        if 'element_electron_shells' in s:
            if 'element_electron_shells' not in ret:
                ret['element_electron_shells'] = []
            ret['element_electron_shells'].extend(s['element_electron_shells'])
        if 'element_ecp' in s:
            if 'element_ecp' in ret:
                raise RuntimeError('Cannot overwrite existing ECP')
            ret['element_ecp'] = s['element_ecp']
            ret['element_ecp_electrons'] = s['element_ecp_electrons']
        if 'element_references' in s:
            if 'element_references' not in ret:
                ret['element_references'] = []
            for ref in s['element_references']:
                if not ref in ret['element_references']:
                    ret['element_references'].append(ref)

    # Sort the shells by angular momentum
    # Note that I don't sort ECP - ECP can't be composed, and
    # it should be sorted in the only source with ECP
    if 'element_electron_shells' in ret:
        ret['element_electron_shells'].sort(key=lambda x: x['shell_angular_momentum'])

    return ret


def prune_basis(basis):
    """
    Removes primitives that have a zero coefficient, and
    removes duplicate shells

    This only finds EXACT duplicates, and is meant to be used
    after uncontracting

    The input basis set is not modified.
    """

    new_basis = copy.deepcopy(basis)

    for k, el in new_basis['basis_set_elements'].items():
        if not 'element_electron_shells' in el:
            continue

        for sh in el['element_electron_shells']:
            new_exponents = []
            new_coefficients = []

            exponents = sh['shell_exponents']

            # transpose of the coefficient matrix
            coeff_t = list(map(list, zip(*sh['shell_coefficients'])))

            # only add if there is a nonzero contraction coefficient
            for i in range(len(sh['shell_exponents'])):
                if not all([float(x) == 0.0 for x in coeff_t[i]]):
                    new_exponents.append(exponents[i])
                    new_coefficients.append(coeff_t[i])

            # take the transpose again, putting the general contraction
            # as the slowest index
            new_coefficients = list(map(list, zip(*new_coefficients)))

            sh['shell_exponents'] = new_exponents
            sh['shell_coefficients'] = new_coefficients

        # Remove any duplicates
        shells = el.pop('element_electron_shells')
        el['element_electron_shells'] = []

        for sh in shells:
            if sh not in el['element_electron_shells']:
                el['element_electron_shells'].append(sh)

    return new_basis


def uncontract_spdf(basis, max_am=0):
    """
    Removes sp, spd, spdf, etc, contractions from a basis set

    The general contractions are replaced by uncontracted versions

    Contractions up to max_am will be left in place. For example,
    if max_am = 1, spd will be split into sp and d

    The input basis set is not modified, and any primitives with all
    zero coefficients are removed
    """

    new_basis = copy.deepcopy(basis)

    for k, el in new_basis['basis_set_elements'].items():

        if not 'element_electron_shells' in el:
            continue
        newshells = []

        for sh in el['element_electron_shells']:

            # am will be a list
            am = sh['shell_angular_momentum']
            coeff = sh['shell_coefficients']

            # if this is an sp, spd,...  orbital
            if len(am) > 1:
                newsh = sh.copy()
                newsh['shell_angular_momentum'] = []
                newsh['shell_coefficients'] = []

                ngen = len(sh['shell_coefficients'])
                for g in range(ngen):
                    if am[g] > max_am:
                        newsh2 = sh.copy()
                        newsh2['shell_angular_momentum'] = [am[g]]
                        newsh2['shell_coefficients'] = [coeff[g]]
                        newshells.append(newsh2)
                    else:
                        newsh['shell_angular_momentum'].append(am[g])
                        newsh['shell_coefficients'].append(coeff[g])

                newshells.insert(0, newsh)

            else:
                newshells.append(sh)

        el['element_electron_shells'] = newshells

    return prune_basis(new_basis)


def uncontract_general(basis):
    """
    Removes the general contractions from a basis set

    The input basis set is not modified, and any primitives with all
    zero coefficients are removed
    """

    new_basis = copy.deepcopy(basis)

    for k, el in new_basis['basis_set_elements'].items():

        if not 'element_electron_shells' in el:
            continue

        newshells = []

        for sh in el['element_electron_shells']:
            # Don't uncontract sp, spd,.... orbitals
            # leave that to uncontract_spdf
            if len(sh['shell_angular_momentum']) == 1:
                for c in sh['shell_coefficients']:
                    # copy, them replace 'shell_coefficients'
                    newsh = sh.copy()
                    newsh['shell_coefficients'] = [c]
                    newshells.append(newsh)
            else:
                newshells.append(sh)

        el['element_electron_shells'] = newshells

    return prune_basis(new_basis)


def uncontract_segmented(basis):
    """
    Removes the segmented contractions from a basis set

    This implicitly removes general contractions as well,
    but will leave sp, spd, ... orbitals alone

    The input basis set is not modified.
    """

    new_basis = copy.deepcopy(basis)

    for k, el in new_basis['basis_set_elements'].items():

        if not 'element_electron_shells' in el:
            continue

        newshells = []

        for sh in el['element_electron_shells']:
            exponents = sh['shell_exponents']
            nam = len(sh['shell_angular_momentum'])

            for i in range(len(exponents)):
                newsh = sh.copy()
                newsh['shell_exponents'] = [exponents[i]]
                newsh['shell_coefficients'] = [["1.00000000"] * nam]

                # Remember to transpose the coefficients
                newsh['shell_coefficients'] = list(map(list, zip(*newsh['shell_coefficients'])))

                newshells.append(newsh)

        el['element_electron_shells'] = newshells

    return new_basis


def make_general(basis):
    """
    Makes one large general contraction for each angular momentum

    If split_spdf is True, sp... orbitals will be split apary
    """

    zero = '0.00000000'

    new_basis = uncontract_spdf(basis)

    for k, el in new_basis['basis_set_elements'].items():
        if not 'element_electron_shells' in el:
            continue

        # See what we have
        all_am = []
        for sh in el['element_electron_shells']:
            if not sh['shell_angular_momentum'] in all_am:
                all_am.append(sh['shell_angular_momentum'])

        all_am = sorted(all_am)

        newshells = []
        for am in all_am:
            # TODO - Check all shells to make sure region and harmonic type are consistent
            newsh = {
                'shell_angular_momentum': am,
                'shell_exponents': [],
                'shell_coefficients': [],
                'shell_region': 'combined',
                'shell_harmonic_type': 'spherical'
            }

            # Do exponents first
            for sh in el['element_electron_shells']:
                if sh['shell_angular_momentum'] != am:
                    continue
                newsh['shell_exponents'].extend(sh['shell_exponents'])

            # Number of primitives in the new shell
            nprim = len(newsh['shell_exponents'])

            cur_prim = 0
            for sh in el['element_electron_shells']:
                if sh['shell_angular_momentum'] != am:
                    continue

                ngen = len(sh['shell_coefficients'])

                for g in range(ngen):
                    coef = [zero] * cur_prim
                    coef.extend(sh['shell_coefficients'][g])
                    coef.extend([zero] * (nprim - len(coef)))
                    newsh['shell_coefficients'].append(coef)

                cur_prim += len(sh['shell_exponents'])

            newshells.append(newsh)

        el['element_electron_shells'] = newshells

    return prune_basis(new_basis)


def _is_single_column(col):
    return sum(float(x) != 0.0 for x in col) == 1


def _is_zero_column(col):
    return sum(float(x) != 0.0 for x in col) == 0


def _nonzero_range(vec):
    for idx, x in enumerate(vec):
        if float(x) != 0.0:
            first = idx
            break

    for idx, x in enumerate(reversed(vec)):
        if float(x) != 0.0:
            last = (len(vec) - idx)
            break

    if first is None:
        return (None, None)
    else:
        return (first, last)


def _find_block(mat):

    # Initial range of rows
    row_range = _nonzero_range(mat[0])
    rows = range(row_range[0], row_range[1])

    # Find the right-most column with a nonzero in it
    col_range = (0, 0)
    for r in rows:
        x, y = _nonzero_range([col[r] for col in mat])
        col_range = (min(col_range[0], x), max(col_range[1], y))

    cols = range(col_range[0], col_range[1])

    # Columns may be jagged also
    # Iterate until we don't see any change
    while True:
        row_range_old = row_range
        col_range_old = col_range
        for c in cols:
            x, y = _nonzero_range(mat[c])
            row_range = (min(row_range[0], x), max(row_range[1], y))

        rows = range(row_range[0], row_range[1])

        for r in rows:
            x, y = _nonzero_range([col[r] for col in mat])
            col_range = (min(col_range[0], x), max(col_range[1], y))

        cols = range(col_range[0], col_range[1])

        if col_range == col_range_old and row_range == row_range_old:
            break

    return (rows, cols)


def optimize_general(basis):
    """
    Optimizes the general contraction using the method of Hashimoto et al

    .. seealso :: | T. Hashimoto, K. Hirao, H. Tatewaki
                  | 'Comment on Dunning's correlation-consistent basis set'
                  | Chemical Physics Letters v243, Issues 1-2, pp, 190-192 (1995)
                  | https://doi.org/10.1016/0009-2614(95)00807-G

    """

    new_basis = copy.deepcopy(basis)
    for k, el in new_basis['basis_set_elements'].items():

        if not 'element_electron_shells' in el:
            continue

        elshells = el.pop('element_electron_shells')
        el['element_electron_shells'] = []
        for sh in elshells:
            exponents = sh['shell_exponents']
            coefficients = sh['shell_coefficients']
            nprim = len(exponents)
            nam = len(sh['shell_angular_momentum'])

            if nam > 1 or len(coefficients) < 2:
                el['element_electron_shells'].append(sh)
                continue

            # First, find columns (general contractions) with a single non-zero value
            single_columns = [idx for idx, c in enumerate(coefficients) if _is_single_column(c)]

            # Find the corresponding rows that have a value in one of these columns
            # Note that at this stage, the row may have coefficients in more than one
            # column. That is ok, we are going to split it off anyway
            single_rows = []
            for col_idx in single_columns:
                col = coefficients[col_idx]
                for row_idx in range(nprim):
                    if float(col[row_idx]) != 0.0:
                        single_rows.append(row_idx)

            # Split those out into new shells, and remove them from the
            # original shell
            new_shells_single = []
            for row_idx in single_rows:
                newsh = copy.deepcopy(sh)
                newsh['shell_exponents'] = [exponents[row_idx]]
                newsh['shell_coefficients'] = [['1.00000000000']]
                new_shells_single.append(newsh)

            exponents = [x for idx, x in enumerate(exponents) if idx not in single_rows]
            coefficients = [x for idx, x in enumerate(coefficients) if idx not in single_columns]
            coefficients = [[x for idx, x in enumerate(col) if not idx in single_rows] for col in coefficients]

            # Remove Zero columns
            #coefficients = [ x for x in coefficients if not _is_zero_column(x) ]

            # Find contiguous rectanglar blocks
            new_shells = []
            while len(exponents) > 0:
                block_rows, block_cols = _find_block(coefficients)

                # add as a new shell
                newsh = copy.deepcopy(sh)
                newsh['shell_exponents'] = [exponents[i] for i in block_rows]
                newsh['shell_coefficients'] = [[coefficients[colidx][i] for i in block_rows] for colidx in block_cols]
                new_shells.append(newsh)

                # Remove from the original exponent/coefficient set
                exponents = [x for idx, x in enumerate(exponents) if idx not in block_rows]
                coefficients = [x for idx, x in enumerate(coefficients) if idx not in block_cols]
                coefficients = [[x for idx, x in enumerate(col) if not idx in block_rows] for col in coefficients]

            # I do this order to mimic the output of the original BSE
            el['element_electron_shells'].extend(new_shells)
            el['element_electron_shells'].extend(new_shells_single)

        # Fix coefficients for completely uncontracted shells to 1.0
        for sh in el['element_electron_shells']:
            if len(sh['shell_coefficients']) == 1 and len(sh['shell_coefficients'][0]) == 1:
                sh['shell_coefficients'][0][0] = '1.0000000'

    return new_basis


def sort_shells(shells):
    """
    Sort a list of basis set shells into a standard order

    The order within a shell is by decreasing value of the exponent.

    The order of the shell list is in increasing angular momentum, and then
    by decreasing number of primitives, then decreasing value of the largest exponent.

    The original data is not modified.
    """

    new_shells = copy.deepcopy(shells)

    for sh in new_shells:
        # Sort primitives within a shell
        # Transpose of coefficients
        tmp_c = list(map(list, zip(*sh['shell_coefficients'])))

        # Zip together exponents and coeffs for sorting
        tmp = zip(sh['shell_exponents'], tmp_c)

        # Sort by decreasing value of exponent
        tmp = sorted(tmp, key=lambda x: -float(x[0]))

        # Unpack, and re-transpose the coefficients
        tmp_c = [x[1] for x in tmp]
        sh['shell_exponents'] = [x[0] for x in tmp]
        sh['shell_coefficients'] = list(map(list, zip(*tmp_c)))

    # Sort by increasing AM, then general contraction level, then decreasing highest exponent
    return list(
        sorted(
            new_shells,
            key=
            lambda x: (max(x['shell_angular_momentum']), -len(x['shell_exponents']), -len(x['shell_coefficients']), -float(x['shell_exponents'][0]))
        ))


def sort_potentials(potentials):
    """
    Sort a list of ECP potentials into a standard order

    The order within a potential is not modified.

    The order of the shell list is in increasing angular momentum, with the largest
    angular momentum being moved to the front.

    The original data is not modified, and a deep copy is returned.
    """

    new_potentials = copy.deepcopy(potentials)

    # Sort by increasing AM, then move the last element to the front
    new_potentials = list(sorted(new_potentials, key=lambda x: x['potential_angular_momentum']))
    new_potentials.insert(0, new_potentials.pop())
    return new_potentials


def sort_basis(basis):
    """
    Sorts all the information in a basis set into a standard order

    The original data is not modified.
    """

    new_basis = copy.deepcopy(basis)

    for k, el in new_basis['basis_set_elements'].items():
        if 'element_electron_shells' in el:
            el['element_electron_shells'] = sort_shells(el['element_electron_shells'])

    return new_basis



'''
Some data common to all tests
'''

import basis_set_exchange as bse

# Use random for getting sets of elements
rand_seed = 39466 # from random.org

# Load all the metadata once
bs_metadata = bse.get_metadata()
bs_names = bse.get_all_basis_names()
bs_formats = list(bse.get_formats().keys())
ref_formats = list(bse.get_reference_formats().keys())
true_false = [ True, False ]

# A representative sample of basis sets
bs_names_sample = ['6-31g', '6-31+g*', 'aug-cc-pvtz',
                   'lanl2dz', 'def2-tzvp', 'tzp', 'sto-3g']

# To test role lookup
role_tests = [ ('cc-pvdz', 'rifit', 'cc-pvdz-rifit'),
               ('cc-pvtz', 'rifit', 'cc-pvtz-rifit'),
               ('cc-pvqz', 'rifit', 'cc-pvqz-rifit'),
               ('aug-cc-pvdz', 'rifit', 'aug-cc-pvdz-rifit'),
               ('aug-cc-pvtz', 'rifit', 'aug-cc-pvtz-rifit'),
               ('aug-cc-pvqz', 'rifit', 'aug-cc-pvqz-rifit')
             ]


def bool_matrix(size):
    '''Returns an identity matrix of a given size consisting of bool types
    '''
    ret = [[False for i in range(size)] for j in range(size)]
    for x in range(size):
        ret[x][x] = True
    return ret



'''
Converts basis set data to a specified output format
'''

from collections import OrderedDict
from .bsejson import write_json
from .nwchem import write_nwchem
from .g94 import write_g94
from .gamess_us import write_gamess_us
from .psi4 import write_psi4
from .turbomole import write_turbomole

_converter_map = {
    'json': {
        'display': 'JSON',
        'extension': '.json',
        'comment': None,
        'function': write_json
    },
    'nwchem': {
        'display': 'NWChem',
        'extension': '.nw',
        'comment': '#',
        'function': write_nwchem
    },
    'gaussian94': {
        'display': 'Gaussian94',
        'extension': '.gbs',
        'comment': '!',
        'function': write_g94
    },
    'gamess_us': {
        'display': 'GAMESS US',
        'extension': '.bas',
        'comment': '!',
        'function': write_gamess_us
    },
    'psi4': {
        'display': 'Psi4',
        'extension': '.gbs',
        'comment': '!',
        'function': write_psi4
    },
    'turbomole': {
        'display': 'Turbomole',
        'extension': '.tm',
        'comment': '#',
        'function': write_turbomole
    }
}


def convert_basis(basis_dict, fmt, header=None):
    '''
    Returns the basis set data as a string representing
    the data in the specified output format
    '''

    # make converters case insensitive
    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    ret_str = _converter_map[fmt]['function'](basis_dict)
    if header is not None and fmt != 'json':
        comment_str = _converter_map[fmt]['comment']
        header_str = comment_str + comment_str.join(header.splitlines(True))
        ret_str = header_str + '\n\n' + ret_str

    # HACK - Psi4 requires the first non-comment line be spherical/cartesian
    #        so we have to add that before the header
    if fmt == 'psi4':
        ret_str = basis_dict['basis_set_harmonic_type'] + '\n\n' + ret_str

    return ret_str


def get_formats():
    '''
    Returns the available formats mapped to display name.

    This is returned as an ordered dictionary, with the most common
    at the top, followed by the rest in alphabetical order
    '''

    at_top = ['nwchem', 'gaussian94', 'psi4']
    ret = [(k, v['display']) for k, v in _converter_map.items()]
    ret = OrderedDict(sorted(ret))

    for x in reversed(at_top):
        ret.move_to_end(x, False)

    # Move JSON to the end
    ret.move_to_end('json', True)

    return ret


def get_format_extension(fmt):
    '''
    Returns the recommended extension for a given format
    '''

    fmt = fmt.lower()
    if fmt not in _converter_map:
        raise RuntimeError('Unknown basis set format "{}"'.format(fmt))

    return _converter_map[fmt]['extension']



"""
Tests for sanity of auxiliary basis sets
"""

import basis_set_exchange as bse
import pytest

from .common_testvars import *

@pytest.mark.parametrize('basis_name', bs_names)
def test_aux_sanity(basis_name):
    """For all basis sets, check that

       1. All aux basis sets exist
       2. That the role of the aux basis set matches the role in
          the orbital basis
    """

    this_metadata = bs_metadata[basis_name]

    for role,aux in this_metadata['auxiliaries'].items():
        assert(aux in bs_metadata)
        aux_metadata = bs_metadata[aux]

        assert(role == aux_metadata['role'])

